#include <stdio.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/uio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/wait.h>
#include <sys/mman.h>

#define BUFFER_SIZE 1024


static const int MAX_LINE_SIZE = 1024;
/*
4x3
5 4 4
7 1 7
8 5 4
2 4 4

i * cols + j

5 4 4 7 1 7 8 5 4 2 4 4
 */

void *create_shared_memory(size_t size) {
    int protection = PROT_READ | PROT_WRITE;
    int visibility = MAP_SHARED | MAP_ANONYMOUS;
    return mmap(NULL, size, protection, visibility, -1, 0);
}

int getIndex(int i, int j, int n, int m){
    return i*m+j;
}

int* readMatrix(FILE* file, int* n, int* m){ // n lines ;;;; m cols
    char *str = (char*) malloc(MAX_LINE_SIZE);
    fgets(str, MAX_LINE_SIZE, file);
    char *token = strtok(str,"x");
    *n = atoi(token);
    token = strtok(NULL,"x");
    *m = atoi(token);

    int *matrix = (int *) malloc(*n * *m * sizeof(int));

    for(int i = 0; i < *n; i++){
        fgets(str, MAX_LINE_SIZE, file);
        token = strtok(str," ");
        for (int j = 0; token != NULL && j < *m; j++){
            matrix[getIndex(i,j,*n,*m)] = atoi(token);
            token = strtok(NULL," ");
        }
    }
    return matrix;
}

int main(int argc, char* argv[]) {
    if (argc != 3) {
        printf("usage: addmx file1 file2\n");
        return EXIT_FAILURE;
    }

    FILE *fd = fopen(argv[1], "r");
    FILE *fd2 = fopen(argv[2], "r");

    if (fd == NULL) {
        printf("error: cannot open file1 %s\n", argv[1]);
        return EXIT_FAILURE;
    }

    if (fd2 == NULL) {
        printf("error: cannot open file2 %s\n", argv[2]);
        return EXIT_FAILURE;
    }


    int n, m;
    int n2, m2;
    int *matrix1 = readMatrix(fd, &n, &m);
    int *matrix2 = readMatrix(fd2, &n2, &m2);

    fclose(fd);
    fclose(fd2);

    if (n != n2 || m != m2) {
        printf("Cannot sum matrix1 with different sizes");
        return EXIT_FAILURE;
    }

    int *matrixSum = (int *) create_shared_memory(n * m * sizeof(int));

    pid_t pid[m];
    for (int j = 0; j < m; j++) {  // Cria todos os filhos
        if ((pid[j] = fork()) == -1) {
            perror("fork");
            return EXIT_FAILURE;
        } else if (pid[j] == 0) { // Codigo dos filhos
            for (int i = 0; i < n; i++) {
                int index = getIndex(i, j, n, m);
                matrixSum[index] = matrix1[index] + matrix2[index];
            }
            return EXIT_SUCCESS;  // garantia que os filhos nao avancam daqui
        }
    }

    for (int j = 0; j < m; j++){ // Espera por todos os filhos
        if (waitpid(pid[j], NULL, 0) == -1) {
            perror("wait");
            return EXIT_FAILURE;
        }
    }

    printf("%dx%d\n",n,m);
    for (int i = 0; i < n; i++) {  // Print matrixSum
        for (int j = 0; j < m; j++) {
            if(j == 0){
                printf("%d", matrixSum[getIndex(i, j, n, m)]);
            }else{
                printf(" %d", matrixSum[getIndex(i, j, n, m)]);
            }
        }
        printf("\n");
    }

    return EXIT_SUCCESS;
}